# bear.py

import gzip
import math
import struct
from bisect import bisect_left
from functools import cmp_to_key

DEBUG_PRINT_ = False
ULONG_MAX_ = 0xffffffffffffffff
NUM_US_ = 65536
NUM_NORMALS_ = 2 * 126 * 256 + 1
NUM_ABNORMALS_ = NUM_US_ - NUM_NORMALS_


class Bear:

    def __init__(self, f):
        f.load_label('[Bear]')
        self.num_features = f.load_ulong()
        dprint(f'Number of features: {self.num_features}')
        self.num_labels = f.load_ulong()
        dprint(f'Number of labels: {self.num_labels}')
        self.loss_function_name = f.load_string()
        dprint(f'Loss function: {self.loss_function_name}')
        dprint(f'[Empty model{"" if self.num_labels == 1 else "s"}]')
        self.empty_models = f.load_array(self.num_labels, BearModel)
        dprint(f'[/Empty model{"" if self.num_labels == 1 else "s"}]')

        self.nums_models = f.load_ulong_array(self.num_labels)
        dprint(f'[Numbers of models]')

        for l in range(self.num_labels):
            dprint(f'Label {l}: Number of models: {self.nums_models[l]}')

        dprint(f'[/Numbers of models]')

        self.nums_parents = f.load_ulong_array(self.num_labels)
        dprint(f'[Numbers of parents]')

        for l in range(self.num_labels):
            dprint(f'Label {l}: Number of parents: {self.nums_parents[l]}')

        dprint(f'[/Numbers of parents]')

        self.sums_weights = f.load_double_array(self.num_labels)
        dprint(f'[Sums of weights]')

        for l in range(self.num_labels):
            dprint(f'Label {l}: Sum of weights: {self.sums_weights[l]}')

        dprint(f'[/Sums of weights]')

        self.strengths = f.load_double_array(self.num_labels)
        dprint(f'[Strengths]')

        for l in range(self.num_labels):
            dprint(f'Label {l}: Strength: {self.strengths[l]}')

        dprint(f'[/Strengths]')

        dprint(f'[Models]')
        self.models = f.load_imatrix(
                self.num_labels,
                self.nums_models,
                BearModel)
        dprint(f'[/Models]')

        dprint(f'[Parents]')
        self.parents = f.load_imatrix(
                self.num_labels,
                self.nums_parents,
                BearModel)
        dprint(f'[/Parents]')

        f.load_label('[/Bear]')

        lookups = {}

        for model in self.empty_models:
            lookups[model.id] = model

        for models in self.models + self.parents:
            for model in models:
                lookups[model.id] = model

        for models in self.models + self.parents:
            for model in models:
                model.parent = lookups[model.parent_id]


    def from_filename(filename):
        if not filename.endswith('.bear.gz'):
            filename += '.bear.gz'

        f = ObjectFile(filename)
        self = Bear(f)
        f.close()

        return self


    def predict(self, features):
        if len(features) != self.num_features:
            raise Exception(f'Number of features passed in ({len(features)} ' \
                    + 'does not match number of features in model ' \
                    + f'({self.num_features})')

        predictions = []

        for l in range(self.num_labels):
            if self.num_features > 0 and self.nums_models[l] > 0:
                prediction = 0

                for model in self.models[l]:
                    prediction += model.weight * model.predict(features)

                loss_function = self.empty_models[l].loss_function
                prediction = loss_function.compute_label_estimate(
                        prediction / self.sums_weights[l])
            else:
                prediction = self.empty_models[l].empty_label_estimate

            predictions.append(prediction)

        return predictions


class BearModel:

    def __init__(self, f):
        f.load_label('[BearModel]')
        self.id = f.load_ulong()
        dprint(f'ID: {self.id}')
        self.empty = f.load_bool()
        dprint(f'Empty: {self.empty}')

        if not self.empty:
            self.completeness_trivial = f.load_bool()
            dprint(f'Completeness trivial: {self.completeness_trivial}')
            self.complete_trivial = f.load_bool()
            dprint(f'Complete trivial: {self.complete_trivial}')
            self.incomplete_prediction = f.load_double()
            dprint(f'Incomplete prediction: {self.incomplete_prediction}')
            self.complete_prediction = f.load_double()
            dprint(f'Complete prediction: {self.complete_prediction}')
            self.assembly = BearAssembly(f)
            self.weight = f.load_double()
            dprint(f'Weight: {self.weight}')
            self.parent_id = f.load_ulong()
            dprint(f'Parent ID: {self.parent_id}')

        self.min_input = f.load_double()
        dprint(f'Min input: {self.min_input}')
        self.max_input = f.load_double()
        dprint(f'Max input: {self.max_input}')

        if self.empty:
            self.loss_function_name = f.load_string()
            dprint(f'Loss function: {self.loss_function_name}')
            self.loss_function = LossFunction(self.loss_function_name, f)
            self.empty_label_prediction = f.load_double()
            dprint(f'Empty label prediction: {self.empty_label_prediction}')
            self.empty_label_estimate = f.load_double()
            dprint(f'Empty label estimate: {self.empty_label_estimate}')
        elif not self.complete_trivial:
            self.complete_hypercube = BearHypercube(f)

        f.load_label('[/BearModel]')


    def predict(self, features):
        label = None
        model = self

        while model is not None:
            if model.empty:
                return self.add(label, model.empty_label_prediction, model)

            part = model.assembly.last_part()
            num_features = part.num_features
            complete = True
            part_features = []

            for f in range(num_features):
                f_full = part.features[f]
                feature = features[f_full]

                if feature is None \
                        or paw.double_pawized_is_infinite_or_nan(feature):
                    complete = False
                    break

                part_features.append(feature)

            if complete:
                if not model.complete_trivial:
                    label = self.add(
                            label,
                            model.complete_hypercube.predict(part_features),
                            None)

                label = self.add(
                        label,
                        model.complete_prediction,
                        model)

            else:
                label = self.add(
                        label,
                        model.incomplete_prediction,
                        model)

            model = model.parent


    def add(self, label, prediction, model):
        if prediction == 0:
            if label is None:
                label = 0
            else:
                return label
        else:
            if label is None:
                label = prediction
            else:
                label += prediction

        if model is not None:
            if label < model.min_input:
                label = model.min_input

            if label > model.max_input:
                label = model.max_input

        return label


class BearHypercube:

    def __init__(self, f):
        f.load_label('[BearHypercube]')
        self.num_features = f.load_ulong()
        dprint(f'Number of features: {self.num_features}')
        self.fields = f.load_array(self.num_features, BearField)
        dprint('[Predictions]')
        self.predictions = Hypercube(f)
        dprint('[/Predictions]')
        f.load_label('[/BearHypercube]')


    def predict(self, features):
        o = self.offset(features)
        return self.predictions.get(o)


    def offset(self, features):
        o = 0

        for f in range(self.num_features):
            feature = features[f]
            b = self.fields[f].lookup(feature)
            o += b * self.predictions.strides[f]

        return o


class BearField:

    normals = []
    abnormals = []


    def __init__(self, f):
        f.load_label('[BearField]')
        num_transitions = f.load_ushort()
        dprint(f'Number of transitions: {num_transitions}')
        transitions = f.load_ushort_array(num_transitions)
        self.lookups = [0xffff] * NUM_US_
        i = 0
        dprint('[Lookup transitions]')

        for n in range(NUM_NORMALS_):
            if i < num_transitions and n == transitions[i]:
                dprint(f'{i} -> {i + 1}: {self.normals[n].to_double():.4g}')
                i += 1

            self.lookups[self.normals[n].u] = i

        dprint('[/Lookup transitions]')
        f.load_label('[/BearField]')


    def lookup(self, value):
        return self.lookups[paw(value).u]


class Hypercube:

    def __init__(self, f):
        f.load_label('[Hypercube]')
        self.type = f.load_string()
        dprint(f'Type: {self.type}')

        if self.type != 'double':
            raise Exception('Only double Hypercubes currently implemented')

        self.size = f.load_ulong()
        dprint(f'Bytes per element: {self.size}')
        self.num_dimensions = f.load_ulong()
        dprint(f'Number of dimensions: {self.num_dimensions}')
        self.dimensions = f.load_ulong_array(self.num_dimensions)

        dprint('[Dimensions]')
        self.num_elements = 1

        for d in range(self.num_dimensions):
            dprint(f'Dimension {d}: {self.dimensions[d]}')
            self.num_elements *= self.dimensions[d]

        dprint('[/Dimensions]')
        dprint(f'Number of elements: {self.num_elements}')

        self.strides = [None] * self.num_dimensions
        stride = 1

        for d_rev in range(self.num_dimensions):
            d = self.num_dimensions - 1 - d_rev
            self.strides[d] = stride
            stride *= self.dimensions[d]

        dprint('[Strides]')

        for d in range(self.num_dimensions):
            dprint(f'Stride {d}: {self.strides[d]}')

        dprint('[/Strides]')

        self.data = f.load_double_array(self.num_elements)

        dprint('[Contents]')
        self.print(0, 0)
        dprint('[/Contents]')

        f.load_label('[/Hypercube]')


    def print(self, d, s):
        if self.num_dimensions == 0:
            dprint(f'{self.data[0]}')
            return

        dimension = self.dimensions[d]

        for i in range(dimension):
            o = s + self.strides[d] * i
            dprint(f'[i{d} = {i}]')

            if d == self.num_dimensions - 1:
                dprint(f'{self.data[o]}')
            else:
                self.print(d + 1, o)

            dprint(f'[/i{d} = {i}]')

    def get(self, o):
        return self.data[o]


class paw:

    def __init__(self, d):
        bs = struct.pack('<d', d)
        dm = bs[0] | bs[1] << 8 | bs[2] << 16 | bs[3] << 24 | bs[4] << 32 \
                | bs[5] << 40 | (bs[6] & 0x0f) << 48
        e = (bs[6] & 0xf0) >> 4 | (bs[7] & 0x7f) << 4
        n = (bs[7] & 0x80) >> 7

        if e == 0:
            self.u = 0
            return

        if e == 0x7ff:
            if dm == 0:
                self.u = 0xff00 if n else 0x7f00
            else:
                self.u = 0x7f01

            return

        m = dm >> 44
        r = dm & 0xfffffffffff
        s = 0x80000000000

        if r > s or r == s and m & 0x1 == 0x1:
            m += 1

            if m == 0x100:
                e += 1
                m = 0

        if e > 1086:
            self.u = 0xff00 if n else 0x7f00
            return

        if e < 961:
            self.u = 0
            return

        e -= 960
        self.u = m | e << 8 | n << 15


    def to_double(self):
        if self.u == 0:
            return 0

        bs = [0, 0, 0, 0, 0]
        bs += [(self.u & 0x000f) << 4]
        e = ((self.u & 0x7f00) >> 8) + 960
        bs += [(self.u & 0x00f0) >> 4 | (e & 0x00f) << 4]
        bs += [(e & 0xff0) >> 4 | (self.u & 0x8000) >> 8]

        return struct.unpack('<d', bytes(bs))[0]


    def from_u(u):
        self = paw(0)
        self.u = u
        return self


    def is_infinite_or_nan(self):
        return self.u & 0x7f00 == 0x7f00


    def double_pawized_is_infinite_or_nan(d):
        return paw(d).is_infinite_or_nan()


class BearAssembly:

    def __init__(self, f):
        f.load_label('[BearAssembly]')
        self.num_parts = f.load_ulong()
        dprint(f'Number of parts: {self.num_parts}')
        self.parts = f.load_array(self.num_parts, BearPart)
        f.load_label('[/BearAssembly]')


    def last_part(self):
        return self.parts[self.num_parts - 1]


class BearPart:

    def __init__(self, f):
        f.load_label('[BearPart]')
        self.num_features = f.load_ulong()
        dprint(f'Number of features: {self.num_features}')
        self.features = f.load_ulong_array(self.num_features)

        dprint(f'[Features]')

        for ff in range(self.num_features):
            dprint(f'Feature {ff}: {self.features[ff]}')

        dprint(f'[/Features]')

        f.load_label('[/BearPart]')


class LossFunction:

    def __init__(self, name, f):
        self.name = name

        if name == 'MSE':
            pass

        elif name == 'MAE':

            if f.load_bool():
                raise Exception('Impossible: do_labels cannot be true')

            self.ranker = Ranker(f)

        elif name == 'LOG':
            self.min_label_prediction_or_estimate = f.load_double()
            dprint(f'Minimum label prediction or estimate: ' +
                  '{self.min_label_prediction_or_estimate}')
            self.max_label_prediction_or_estimate = f.load_double()
            dprint(f'Maximum label prediction or estimate: ' +
                  '{self.max_label_prediction_or_estimate}')

        elif name == 'BALANCED_LOG':
            self.min_label_prediction = f.load_double()
            dprint(f'Minimum label prediction: {self.min_label_prediction}')
            self.max_label_prediction = f.load_double()
            dprint(f'Maximum label prediction: {self.max_label_prediction}')
            self.p = f.load_double()
            dprint(f'P: {self.p}')
            self.max_label_estimate = None
            self.min_label_estimate \
                    = self.bl_compute_label_estimate(self.min_label_prediction)
            self.max_label_estimate \
                    = self.bl_compute_label_estimate(self.max_label_prediction)

        else:
            raise Exception(f"Loss function name '{name}' not recognized")


    def compute_label_estimate(self, label_prediction):
        if self.name == 'MSE':
            return label_prediction

        if self.name == 'MAE':
            return self.ranker.value_from_crank(label_prediction)

        if self.name == 'LOG':
            if label_prediction < 0 or label_prediction > 1:
                raise Exception(f'Label prediction {label_prediction} ' +
                                'must be in [0, 1]')

            if label_prediction < self.min_label_prediction_or_estimate:
                return self.min_label_prediction_or_estimate

            if label_prediction > self.max_label_prediction_or_estimate:
                return self.max_label_prediction_or_estimate

            return label_prediction

        if self.name == 'BALANCED_LOG':
            if label_prediction < 0 or label_prediction > 1:
                raise Exception(f'Label prediction {label_prediction} ' +
                                'must be in [0, 1]')

            if label_prediction < self.min_label_prediction:
                label_prediction = self.min_label_prediction

            if label_prediction > self.max_label_prediction:
                label_prediction = self.max_label_prediction

            x = (1 - self.p) * label_prediction
            label_estimate = x / (p * (1 - label_prediction) + x)

            if self.max_label_estimate is not None and (
                    label_estimate < self.min_label_estimate
                    or label_estimate > self.max_label_estimate):
                raise Exception(f'Impossible: {label_estimate} should be' \
                        + f'in [{self.min_label_estimate}, '
                        + f'{self.max_label_estimate}]')

            return label_estimate

        raise Exception('Impossible')


    def bl_compute_label_estimate(self, label_prediction):

        if label_prediction < 0 or label_prediction > 1:
            raise Exception(
                    f'Label prediction {label_prediction} must be in [0,1]')

        if label_prediction < self.min_label_estimate:
            label_prediction = self.min_label_prediction

        if label_prediction > self.max_label_estimate:
            label_prediction = self.max_label_prediction

        x = (1 - self.p) * label_prediction
        label_estimate = x / (self.p * (1 - label_prediction) + x)

        if self.max_label_estimate is not None \
                and (label_estimate < self.min_label_estimate \
                or label_estimate > self.max_label_estimate):
            raise Exception(
                    'Impossible: {label_estimate} should be in ' +
                    '[{self.min_label_estimate}, {self.max_label_estimate}]')

        return label_estimate


def compare(a, b):
    if len(a) == 1 and len(b) == 3:
        needle = a
        hay = b
        s = +1
    elif len(a) == 3 and len(b) == 1:
        needle = b
        hay = a
        s = -1
    else:
        raise Exception(f'Cannot understand {a} and {b}')

    needle_rank = needle[0]

    if needle_rank < hay[1]:
        return -s

    if needle_rank < hay[2]:
        return 0

    return s


class Ranker:

    def __init__(self, f):
        f.load_label('[Ranker]')
        self.total_frequency = f.load_ulong()
        dprint(f'Total frequency: {self.total_frequency}')
        self.array = KeyValueArray(f)
        self.min_value = self.array.get(0)[1]
        self.max_value = self.array.get(self.array.num_pairs - 1)[1]
        f.load_label('[/Ranker]')


    def value_from_crank(self, crank):
        if crank <= -self.total_frequency:
            return self.min_value

        if crank >= self.total_frequency:
            return self.max_value

        rank = (crank + self.total_frequency) / 2

        needle = [int(math.floor(rank))]
        i = bisect_left(
                self.array.data,
                cmp_to_key(compare)(needle),
                key=cmp_to_key(compare))

        if i == len(self.array.data):
            raise Exception(f'Could not find rank {rank}')

        value = self.array.data[i][0]

        if rank == needle[0] and rank > 0:
            needle[0] -= 1
            i = bisect_left(
                    self.array.data,
                    cmp_to_key(compare)(needle),
                    key=cmp_to_key(compare))

            if i == len(self.array.data):
                raise Exception(f'Could not find rank {rank}')

            value = (value + self.array.data[i][0]) / 2

        return value


class KeyValueArray:

    def __init__(self, f):
        f.load_label('[KeyValueArray]')
        self.num_pairs = f.load_ulong()
        dprint(f'Number of pairs: {self.num_pairs}')
        self.info = KeyValueInfo(f)

        if self.info.num_keys != 1 or self.info.max_keys[0] != ULONG_MAX_ \
                or self.info.num_values != 1 \
                or self.info.max_values[0] != ULONG_MAX_:
            raise Exception('KeyValueInfo specified not supported')

        self.data = f.load_ranker_entry_array(self.num_pairs)
        dprint('[Entries]')

        for i in range(len(self.data)):
            entry = self.data[i]

            if i < len(self.data) - 1:
                entry[2] = self.data[i + 1][1]

            dprint(f'[Entry {i}]')
            dprint(f'Value: {entry[0]}')
            dprint(f'Centered rank: {entry[1]}')
            dprint(f'Next value: {entry[2]}')
            dprint(f'[/Entry {i}]')

        dprint('[/Entries]')
        f.load_label('[/KeyValueArray]')


    def get(self, i):
        return self.data[i]


class KeyValueInfo:

    def __init__(self, f):
        f.load_label('[KeyValueInfo]')
        self.num_keys = f.load_ulong()
        dprint(f'Number of keys: {self.num_keys}')

        self.max_keys = f.load_ulong_array(self.num_keys)
        dprint(f'[Maximum keys]')

        for k in range(self.num_keys):
            dprint(f'Key {k}: Maximum key: {self.max_keys[k]}')

        dprint(f'[/Maximum keys]')

        self.num_values = f.load_ulong()
        dprint(f'Number of values: {self.num_values}')

        self.max_values = f.load_ulong_array(self.num_values)
        dprint(f'[Maximum values]')

        for v in range(self.num_values):
            dprint(f'Value {k}: Maximum value: {self.max_values[v]}')

        dprint(f'[/Maximum values]')

        f.load_label('[/KeyValueInfo]')


class ObjectFile:

    def __init__(self, filename):
        self.f = gzip.open(filename, 'rb')


    def close(self):
        self.f.close()


    def read(self, n):
        return self.f.read(n)


    def load_bool(self):
        return bool.from_bytes(self.read(1), byteorder='little', signed=False)


    def load_ulong(self):
        return int.from_bytes(self.read(8), byteorder='little', signed=False)


    def load_ushort(self):
        return int.from_bytes(self.read(2), byteorder='little', signed=False)


    def load_ranker_entry(self):
        return [self.load_double(), self.load_ulong(), 0]


    def load_double(self):
        return struct.unpack('<d', self.read(8))[0]


    def load_label(self, label):
        e = len(label)
        l = self.load_ulong()

        if l != e:
            raise Exception(f'Expected label length {e} for "{label}" ' +
                            f'but read length {l}')

        r = str(self.read(e), 'utf-8')

        if r != label:
            raise Exception(f'Expected label {label} but read {r}')

        dprint(label)


    def load_string(self):
        return str(self.read(self.load_ulong()), 'utf-8')


    def load_array(self, n, cls):
        return [cls(self) for i in range(n)]


    def load_imatrix(self, n0, n1s, cls):
        return [[cls(self) for i1 in range(n1s[i0])] for i0 in range(n0)]


    def load_ushort_array(self, n):
        return [self.load_ushort() for i in range(n)]


    def load_ulong_array(self, n):
        return [self.load_ulong() for i in range(n)]


    def load_double_array(self, n):
        return [self.load_double() for i in range(n)]


    def load_ranker_entry_array(self, n):
        return [self.load_ranker_entry() for i in range(n)]


indent_ = 0


def dprint(s):

    global indent_

    if DEBUG_PRINT_:

        if s[0:2] == '[/':
            indent_ -= 1

        prefix = ''

        for i in range(indent_):
            prefix += '. '

        print(prefix + s)

        if s[0:2] != '[/' and s[0] == '[':
            indent_ += 1


# Static initialization.

for u in range(NUM_US_):
    if u == 0 or (u & 0x7f00 != 0 and u & 0x7f00 != 0x7f00):
        BearField.normals.append(paw.from_u(u))
    else:
        BearField.abnormals.append(u)

if len(BearField.normals) != NUM_NORMALS_:
    raise Exception(f'Impossible: {len(BearField.normals)} v {NUM_NORMALS_}')

if len(BearField.abnormals) != NUM_ABNORMALS_:
    raise Exception(
            f'Impossible: {len(BearField.abnormals)} v {NUM_ABNORMALS_}')

BearField.normals.sort(key=paw.to_double)

for u in range(NUM_NORMALS_ - 1):
    if BearField.normals[u].to_double() \
            >= BearField.normals[u + 1].to_double():
        raise Exception('Impossible')
