#!/usr/bin/env python3
import argparse
import csv
import json
import pathlib
import os
import sys
from bear import Bear


def detect_filetype(filename):
    filetype = pathlib.Path(filename).suffix[1:].lower()

    if filetype == '':
        raise Exception('Could not infer a filetype extension from the ' \
                + f"filename '{filename}'")

    if filetype not in ['csv', 'tsv', 'hsv', 'nsv', 'txt']:
        raise Exception('Do not recognize the filetype extension ' \
                + f"'{filetype}'")

    return filetype


def separator(t):
    return {'csv': ',', 'tsv': '\t', 'hsv': '\1', 'nsv': '\0', 'txt': '\0'}[t]


def veto(args, columns):
    n = 0

    for c in columns:
        if args[c] is not None:
            n += 1

    if n > 1:
        raise Exception(f'Can only specify one of {columns}')


def find(header, column_names):
    if type(column_names) == list:
        columns = []

        for n in column_names:
            columns.append(find(header, n))

        return columns

    for c in range(len(header)):
        if header[c] == column_names:
            return c

    raise Exception(f"Did not find column name '{column_names}' in the header")


parser = argparse.ArgumentParser(
        prog='bear_predict',
        description='Predict by streaming examples through a pre-trained ' + \
                'Bear model')
parser.add_argument(
        'input_filename',
        metavar='input-filename',
        default=sys.stdin)
parser.add_argument(
        'bear_filename',
        metavar='bear-filename')
parser.add_argument(
        'output_filename',
        metavar='output-filename',
        default=sys.stdout)
parser.add_argument(
        '-l',
        '--label-column',
        type=int,
        help='index of a label column to skip')
parser.add_argument(
        '-m',
        '--multi-label-columns',
        help='indexes of label columns to skip, as JSON list')
parser.add_argument(
        '-i',
        '--id-column',
        type=int,
        help='index of an ID column to pass through')
parser.add_argument(
        '-j',
        '--multi-id-columns',
        help='indexes of ID columns to pass through')
parser.add_argument(
        '-f',
        '--has-frequency-column',
        action='store_true',
        help='One column is a frequency column to skip')
parser.add_argument(
        '-c',
        '--frequency-column',
        type=int,
        help='Frequency column')
parser.add_argument(
        '-H',
        '--has-header-row',
        action='store_true',
        help='The first row is a header row')
parser.add_argument(
        '-y',
        '--confirm-header-row',
        action='store_true',
        help='Confirm that a header row of numbers is correct')
parser.add_argument(
        '-L',
        '--label-name',
        help='Name of a label column to skip')
parser.add_argument(
        '-M',
        '--multi-label-names',
        help='iNames of label columns to skip, as JSON list')
parser.add_argument(
        '-I',
        '--id-name',
        help='Name of an ID column to pass through')
parser.add_argument(
        '-J',
        '--multi-id-names',
        help='Names of ID columns to pass through')
parser.add_argument(
        '-C',
        '--frequency-name',
        help='Name of the frequency column to skip')
parser.add_argument(
        '-e',
        '--missing-string',
        default='',
        help="String for a missing value (default = '')")
parser.add_argument(
        '-t',
        '--input-filetype',
        help='Type of the input file')
parser.add_argument(
        '-T',
        '--output-filetype',
        help='Type of the output file')
parser.add_argument(
        '-x',
        '--no-features-out',
        action='store_true',
        help="Don't write features to the output file")

args = vars(parser.parse_args())
print('Welcome to Bear Machine Learning!')
input_filename = args['input_filename']
print(f"Input filename: '{input_filename}'")
print('Input filetype: ' + (f"'{args['input_filetype']}'" \
        if args['input_filetype'] else '(auto-detect)'))
in_sep = separator(args['input_filetype'] or detect_filetype(input_filename))
bear_filename = args['bear_filename']
print(f"Bear model filename: '{bear_filename}'")
output_filename = args['output_filename']
print(f"Output filename: '{output_filename}'")
print('Output filetype: ' + (f"'{args['output_filetype']}'" \
        if args['output_filetype'] else '(auto-detect)'))
out_sep = separator(args['output_filetype'] \
        or detect_filetype(output_filename))
has_header_row = args['has_header_row']
confirm_header_row = args['confirm_header_row']
has_frequency_column = args['has_frequency_column']
missing_string = args['missing_string'] or ''
no_features_out = args['no_features_out']

bear = Bear.from_filename(bear_filename)
num_features = bear.num_features

with (sys.stdin if input_filename == '-'
      else open(input_filename, newline='')) as infile:
    reader = csv.reader(infile, delimiter=in_sep)

    if has_header_row:
        header = next(reader)

    veto(args, [
        'label_column',
        'label_name',
        'multi_label_columns',
        'multi_label_names'])

    if args['label_column'] is not None:
        label_columns = [args['label_column']]
    elif args['label_name'] is not None:
        label_columns = [find(header, args['label_name'])]
    elif args['multi_label_columns'] is not None:
        label_columns = json.loads(args['multi_label_columns'])
    elif args['multi_label_names'] is not None:
        label_columns = find(header, json.loads(args['multi_label_names']))
    else:
        label_columns = []

    if len(label_columns) != len(set(label_columns)):
        raise Exception(f'Repeated label column in {label_columns}')

    if len(label_columns) > 0:
        print(f'Label columns: {label_columns}')

    veto(args, ['id_column', 'id_name', 'multi_id_columns', 'multi_id_names'])

    if args['id_column'] is not None:
        id_columns = [args['id_column']]
    elif args['id_name'] is not None:
        id_columns = [find(header, args['id_name'])]
    elif args['multi_id_columns'] is not None:
        id_columns = json.loads(args['multi_id_columns'])
    elif args['multi_id_names'] is not None:
        id_columns = find(header, json.loads(args['multi_id_names']))
    else:
        id_columns = []

    if len(id_columns) != len(set(id_columns)):
        raise Exception(f'Repeated id column in {id_columns}')

    if len(id_columns) > 0:
        print(f'ID columns: {id_columns}')

    veto(args, ['frequency_column', 'frequency_name'])

    if has_frequency_column:
        if args['frequency_column'] is not None:
            frequency_columns = [args['frequency_column']]
        elif args['frequency_name'] is not None:
            frequency_columns = [find(header, args['frequency_name'])]
        else:
            raise('Frequency column or name not supplied')
    else:
         frequency_columns = []

    if len(frequency_columns) > 0:
        print(f'Frequency column: {frequency_columns[0]}')

    num_columns = num_features + len(label_columns) + len(id_columns) \
            + len(frequency_columns)

    all_columns = label_columns + id_columns + frequency_columns
    feature_columns = []

    for c in range(num_columns):
        if c not in all_columns:
            feature_columns.append(c)

    if len(feature_columns) != num_features:
        raise Exception(
                f'Number of feature columns should be {num_features}; ' \
                + f'{feature_columns} has {len(feature_columns)}')

    all_columns.extend(feature_columns)

    if len(all_columns) != len(set(all_columns)):
        raise Exception(f'Repeated column in {all_columns}')

    for c in all_columns:
        if c >= len(all_columns):
            raise Exception(f'Invalid column {c} specified: leaves a gap')

    with (sys.stdout if output_filename == '-'
          else open(output_filename, 'w')) as outfile:
        writer = csv.writer(
                outfile,
                delimiter=out_sep,
                lineterminator=os.linesep)

        if has_header_row:
            orow = []

            for c in id_columns:
                orow.append(header[c] if c < len(header) else f'id{c}')

            if not no_features_out:
                for c in feature_columns:
                    orow.append(header[c] if c < len(header) else f'f{c}')

            for it in range(2 if len(label_columns) > 0 else 1):
                prefix = 'predicted' if it == 0 else 'actual'

                for c in label_columns:
                    orow.append(f'{prefix}_' + \
                            (header[c] if c < len(header) else f'l{c}'))

            writer.writerow(orow)

        r = 0

        for row in reader:
            orow = []

            for c in id_columns:
                if c >= len(row) or row[c] == missing_string:
                    id = None
                else:
                    id = row[c]

                orow.append(id)

            features = []

            for c in feature_columns:
                if c >= len(row) or row[c] == missing_string:
                    feature = None
                else:
                    feature = float(row[c])

                features.append(feature)

            if not no_features_out:
                orow.extend(features)

            labels = []

            for c in label_columns:
                if c >= len(row) or row[c] == missing_string:
                    label = None
                else:
                    label = float(row[c])

                labels.append(label)

            predictions = bear.predict(features)
            orow.extend(predictions)
            orow.extend(labels)

            writer.writerow(orow)
            r += 1

            if r % 10000 == 0:
                print(f'{r} data rows processed ...')

        print(f'{r} data rows processed in total.')
        print('Thank you for using Bear Machine Learning!')
